/*
 * for the Matrox Marvel G200 and Rainbow Runner-G series
 *
 * This module is an interface to the KS0127 video decoder chip.
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 *****************************************************************************
 */
#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <asm/io.h>

#include "zr36060.h"

#define dprintk     if (debug) printk

#define XGENIOCTRL      0x2a
#define XGENIODATA      0x2b

#define VI_IVSYNC    0x0001 /* input vsync interrupt */
#define VI_CMDCMPL   0x0002 /* command complete interrupt */
#define VI_BLVL      0x0004 /* buffer level interrupt */
#define VI_EOI       0x0008 /* codec end of image interrupt */

static int debug = 1; /* insmod parameter */


/****************************************************************************
* linux kernel module api
****************************************************************************/
#ifdef MODULE
int init_module(void)
#else
int zr36060_init(void)
#endif
{

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
}
#endif


static void zr36060_misc(struct zr36060 *zr, int value)
{
	int	codecctl;
	int	misc;

	codecctl = (zr->codecctl_shadow | 0x01) & ~0x04;

	writeb(codecctl, zr->codecctl);
	writeb(codecctl, zr->codecctl);
	
	value &= 0xff;
	misc = ~((value & 0x9f) | (~value & 0x60)) & 0xff;
	zr->codecmis_shadow = value & 0xff;

	writeb(misc, zr->codecctl + 3);

	writeb(zr->codecctl_shadow, zr->codecctl);

printk("codecctl-misc = 0x%x->x%x\n", value, misc);
}

static void zr36060_and_or_codecctl(struct zr36060 *zr, u16 and, u16 or)
{
	int	codecctl;

	codecctl = (zr->codecctl_shadow & and) | or;
	zr->codecctl_shadow = codecctl;

	writew(codecctl, zr->codecctl);
}


static int
zr36060_read_buffer(struct zr36060 *zr, int start_reg, char *buffer, int length)
{
	int	i;
	u8	*host_ptr;

memset(zr->read_base, 0xff, 512);

	host_ptr = zr->command_base;

	for(i=0; i < length; i++, start_reg++) {
		*host_ptr++ = 2;
		*host_ptr++ = 0;
		*host_ptr++ = 0;
		*host_ptr++ = 0;

		*host_ptr++ = 0;
		*host_ptr++ = 0;
		*host_ptr++ = 0;
		*host_ptr++ = 0;

		*host_ptr++ = 0x01;
		*host_ptr++ = (start_reg >> 8) & 0x03;
		*host_ptr++ = 0x02;
		*host_ptr++ = start_reg & 0xff;

		*host_ptr++ = 0;
		*host_ptr++ = 0;
		*host_ptr++ = 0;
		*host_ptr++ = 0;

		*host_ptr++ = 01;
		*host_ptr++ = 0x8e;
	}

	zr36060_and_or_codecctl(zr, ~0x1, 1);

	/* tell the system where memory buffer for zr36060 exists */
	writel((zr->host_interface_base & ~0x03) | 0x01, zr->codecaddr);


#ifdef	NEVER
	writel(0x0a, (zr->codecctl & ~0xff) | 0x38);
#endif
	writeb(VI_CMDCMPL, zr->viclear);

#ifdef	NEVER
	zr36060_and_or_codecctl(zr, ~0x40, 0x04);
	zr36060_and_or_codecctl(zr, ~0x04, 0x20);
#else
	zr36060_and_or_codecctl(zr, ~0x40, 0x24);
#endif
	zr->codecctl_shadow &= ~0x04;

	for(i=0; i < 0x7ffff; i++) {
		if ( readl(zr->vstatus) & 0x02) {
			writeb(VI_CMDCMPL, zr->viclear);


printk("read_data = ");
for(i=0; i < 16; i++) printk("%x ", zr->read_base[i]); printk("\n");
			return 0;
		}
	}
printk("timeout codec status register = %x\n",readl(zr->vstatus));
printk("read_data = ");
for(i=0; i < 16; i++) printk("%x ", zr->read_base[i]); printk("\n");
	return 1;
}
		
static int
zr36060_write_buffer(struct zr36060 *zr, int start_reg, char *buffer, int length)
{
	u8	*host_ptr;
	int	i;

	if (length > 85) {
		printk("zr36060: Write buffer to large %d starting reg %x\n",
		       length, start_reg);
		return 1;
       }

	writeb(0x00, zr->codecctl);
	writeb(0x03, zr->codecctl);
	writeb(0x03, zr->codecctl);

	/* tell the system where memory buffer for zr36060 exists */
	writel((zr->host_interface_base && ~0x03) | 0x01, zr->codecaddr);

	host_ptr = zr->command_base;

	*host_ptr++ = length * 6;
	*host_ptr++ = 0x02;
	*host_ptr++ = 0;
	*host_ptr++ = 0;
	*host_ptr++ = 0;
	*host_ptr++ = 0;
	*host_ptr++ = 0;
	*host_ptr++ = 0;

	for(i=0; i < length; i++) {
		*host_ptr++ = 1;
		*host_ptr++ = start_reg >> 8;
		*host_ptr++ = 2;
		*host_ptr++ = start_reg;
		*host_ptr++ = 3;
		*host_ptr++ = *buffer++;
	}

	writeb(0x05, zr->codecctl);

	for(i=0; i < 0x7ffff; i++) {
		if ( readl(zr->vstatus) & 0x02) {
printk("command completed\n");
			writeb(0x02, zr->viclear);
			return 0;
		}
	}
printk("timeout codec status register = %x\n", readl(zr->vstatus));

	return 1;
}

static u8 SOF[19] = {
	0xff, 0xc0,		/* Start of frame marker */
	0x00, 0x11,		/* Length of this segment */
	0x08,			/* Precision */
	240/8, 240&0xff,	/* number of lines */
	704/8, 704&0xff,	/* line length */
	3,			/* number of components */
	0, 0x21, 0,		/* Y component */
	1, 0x11, 1,		/* U component */
	2, 0x11, 1		/* V component */
};

static u8 SOS[] = {
	0xff, 0xda,		/* start of Scan marker */
	0x00, 0x0c,		/* lenght of this segment */
	0x03,			/* number of components */
	0x00, 0x00,		/* Y component */
	0x01, 0x11,		/* U component */
	0x02, 0x11,		/* V component */
	0x00, 0x3f, 0x00	/* constant */
};

static u8 DRI[] = {
	0xff, 0xdd,		/* define Restart Interval */
	0x00, 0x04,		/* length of this segment */
	0x00, 0x08,		/* lenght of restart interval in mcu units */
};

static u8 DQT[] = {
	0xff, 0xdb,		/* DHT marker */
	0x00, 0x84,		/* DHT length */
	0x00,			/* table ID 0 */
	0x10, 0x0b, 0x0c, 0x0e, 0x0c, 0x0a, 0x10, 0x0e,
	0x0d, 0x0e, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28,
	0x1a, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23, 0x25,
	0x1d, 0x28, 0x3a, 0x33, 0x3d, 0x3c, 0x39, 0x33,
	0x38, 0x37, 0x40, 0x48, 0x5c, 0x4e, 0x40, 0x44,
	0x57, 0x45, 0x37, 0x38, 0x50, 0x6d, 0x51, 0x57,
	0x5f, 0x62, 0x67, 0x68, 0x67, 0x3e, 0x4d, 0x71,
	0x79, 0x70, 0x64, 0x78, 0x5c, 0x65, 0x67, 0x63,
};

static u8 DQT_2[] = {
	0x01,			/* table ID 1 */
	0x11, 0x12, 0x12, 0x18, 0x15, 0x18, 0x2f, 0x1a,
	0x1a, 0x2f, 0x63, 0x42, 0x38, 0x42, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
	0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63
};

static u8 DHT[] = {
      0xff, 0xdb,		/* DHT marker */
      0x00, 0x84,		/* DHT length */
      0x00,			/* table ID 0 */
      0x10, 0x0b, 0x0c, 0x0e, 0x0c, 0x0a, 0x10, 0x0e,
      0x0d, 0x0e, 0x12, 0x11, 0x10, 0x13, 0x18, 0x28,
      0x1a, 0x18, 0x16, 0x16, 0x18, 0x31, 0x23, 0x25,
      0x1d, 0x28, 0x3a, 0x33, 0x3d, 0x3c, 0x39, 0x33,
      0x38, 0x37, 0x40, 0x48, 0x5c, 0x4e, 0x40, 0x44,
      0x57, 0x45, 0x37, 0x38, 0x50, 0x6d, 0x51, 0x57,
      0x5f, 0x62, 0x67, 0x68, 0x67, 0x3e, 0x4d, 0x71,
      0x79, 0x70, 0x64, 0x78, 0x5c, 0x65, 0x67, 0x63,
};

static u8 DHT_2[] = {
      0x01,			/* table ID 1 */
      0x11, 0x12, 0x12, 0x18, 0x15, 0x18, 0x2f, 0x1a,
      0x1a, 0x2f, 0x63, 0x42, 0x38, 0x42, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63,
      0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63, 0x63
};

static void
zr36060_segments(struct zr36060 *zr, int width, int height)
{
	u8	sof[19];

	/* set jpg mode */

	/* set video mode */

	memcpy(sof, SOF, sizeof(sof));
	sof[5] = width >> 8;
	sof[6] = width;

	sof[7] = height >> 8;
	sof[8] = height;

	zr36060_write_buffer(zr, 0x060, sof, sizeof(sof));
	zr36060_write_buffer(zr, 0x07a, SOS, sizeof(SOS));
	zr36060_write_buffer(zr, 0x0c0, DRI, sizeof(DRI));
	zr36060_write_buffer(zr, 0x0cc, DQT, sizeof(DQT));
	zr36060_write_buffer(zr, 0x0cc+sizeof(DQT), DQT_2, sizeof(DQT_2));
	zr36060_write_buffer(zr, 0x1d4, DHT, sizeof(DHT));
	zr36060_write_buffer(zr, 0x1d4+sizeof(DHT), DHT_2, sizeof(DHT_2));
/*
	zr36060_write_buffer(zr, 0x380, APP, sizeof(APP));
	zr36060_write_buffer(zr, 0x3c0, COM, sizeof(COM));
*/
}

void	zr36060_2A36(struct zr36060 *zr);
void	zr36060_AA17(struct zr36060 *zr);
void	zr36060_EC5B(struct zr36060 *zr, int xc, int x10);
void	zr36060_FB14(struct zr36060 *zr, int xc, int x10);
void	zr36060_FB7C(struct zr36060 *zr);
void	zr36060_FC18(struct zr36060 *zr);
void	zr36060_FC99(struct zr36060 *zr);

void
zr36060_2A36(struct zr36060 *zr)
{
	zr36060_AA17(zr);
	zr36060_FC18(zr);
	zr36060_108D8(zr, 0);
}

void
zr36060_AA17(struct zr36060 *zr)
{
	int	temp_0c;
	int	temp_10;

#ifdef	NEVER
	temp_0c = zr->_28;
#endif
	if (temp_0c != 0) {
		if (temp_0c < 0 || (temp_0c > 2)) {
			zr36060_EC5B(zr, 0x0f, 0);
		} else {
			zr36060_EC5B(zr, 0x05, 0);
		}
	}

	zr36060_EC5B(zr, 0x10, 0);
#ifdef	NEVER
	zr->_71 = 0;
	zr->_6d = zr->_10;
#endif
#ifdef	NEVER
	temp_08 = zr->codecctl_shadow;
	temp_08 &= ~4;
	temp_08 &= ~20;
	temp_08 &= ~40;
	temp_08 &= ~01;
	zr36060_codecctl(zr, temp_08);
#else 
	zr36060_and_or_codecctl(zr, ~0x65, 0);
#endif
	mdelay(100);

#ifdef	NEVER
	temp_08 != 0x01;
	zr36060_codecctl(zr, temp_08);
	zr36060_codecctl(zr, temp_08);
#else
	zr36060_and_or_codecctl(zr, ~0x00, 0x01);
	zr36060_and_or_codecctl(zr, ~0x00, 0x01);
#endif

#ifdef	NEVER
	temp_10 = zr->_28;
#endif
	if (temp_10 != 0) {
		if (temp_10 < 0 || (temp_10 > 2)) {
			zr36060_EC5B(zr, 0x0f, 1);
		} else {
			zr36060_EC5B(zr, 0x05, 1);
		}
	}

	zr36060_EC5B(zr, 0x10, 1);
}

void
zr36060_EC5B(struct zr36060 *zr, int xc, int x10)
{
}


void
zr36060_FB14(struct zr36060 *zr, int xc, int x10)
{
	if (xc & 0xff)
		zr36060_misc(zr, zr->codecmis_shadow & ~0x10);
	else 	
		zr36060_misc(zr, zr->codecmis_shadow | 0x10);
}

void
zr36060_FB7C(struct zr36060 *zr)
{
	int	misc;

	misc = zr->codecmis_shadow;
	mdelay(10);
	zr36060_misc(zr, zr->codecmis_shadow & ~0x04);
	mdelay(1);
	zr36060_misc(zr, zr->codecmis_shadow | 0x04);

	zr36060_misc(zr, misc);
}
	

void
zr36060_FC18(struct zr36060 *zr)
{
	zr36060_FC99(zr);

	zr36060_FB7C(zr);

	zr36060_FC99(zr);
	zr36060_FC99(zr);
	zr36060_FC99(zr);
	zr36060_FC99(zr);
}

void
zr36060_FC99(struct zr36060 *zr)
{
	zr36060_FB14(zr, 0, 0);
	mdelay(100);
	zr36060_misc(zr, zr->codecmis_shadow & ~0x08);
	mdelay(1);
	zr36060_misc(zr, zr->codecmis_shadow | 0x08);
	mdelay(200);
}



int
zr36060_attach(struct zr36060 *zr)
{
	char	buffer[4];

	/* reset the codec interface engine */
	zr36060_and_or_codecctl(zr, 0x00, 0x00);
	zr36060_and_or_codecctl(zr, 0x00, 0x03);
	zr36060_and_or_codecctl(zr, 0x00, 0x03);

	zr36060_misc(zr, 0x98);
	zr36060_FC18(zr);
	zr36060_2A36(zr);

	

	/* tell the system where memory buffer for zr36060 exists */
	writel((zr->host_interface_base & ~0x03) | 0x01, zr->codecaddr);

	/* read id from zr36060 */
	zr36060_read_buffer(zr, 0x22, buffer, 1);
	
#ifdef	NEVER
	zr36060_segments(zr, 704, 240);
	MOD_INC_USE_COUNT;
#endif

	return 0;
}

int
zr36060_deattach(struct zr36060 *zr)
{
#ifdef	NEVER
	/* reset the codec interface */
	writeb(0x00, zr->codecctl);
	writeb(0x01, zr->codecctl);
	writeb(0x00, zr->codecctl);

	MOD_DEC_USE_COUNT;
#endif
	return 0;
}
